import { Application } from 'express';
import { IMountable } from './IMountable';
import { ServeStaticOptions } from 'serve-static';
import { IInitializable } from './IInitializable';
import { TMiddleware } from './TMiddleware';
interface IStaticServe {
    [0]: string;
    [1]?: ServeStaticOptions;
}
interface IMountPoint {
    /**
     * Mount path
     */
    [0]: string;
    /**
     * Mountable object
     */
    [1]: IMountable;
}
export declare class HttpApplication {
    protected readonly expressApp: Application;
    protected readonly mountList: Map<string, Set<IMountPoint>>;
    protected readonly middlewaresBefore: Map<string, Set<TMiddleware<any>>>;
    protected readonly middlewaresAfter: Map<string, Set<TMiddleware<any>>>;
    protected readonly staticList: Map<string, Set<IStaticServe>>;
    protected readonly initializableList: Set<IInitializable>;
    /**
     * @param {e.Application} express
     */
    constructor(express: Application);
    /**
     * Mounts mountable objects to web-path
     * @param {string} path
     * @param {IMountable[]} mountable
     * First element - web-path to which mountable will be mount, second element - mountable object.
     * For example:
     *  ['/test', new TestMountable()]
     * @returns {HttpApplication}
     */
    mount(path: string, mountable: IMountPoint[]): HttpApplication;
    /**
     * Adds middlewares BEFORE all the mountable objects and static-serve middlewares
     * @param {TMiddleware<any>[]} middlewares
     */
    before(middlewares: TMiddleware<any>[]): HttpApplication;
    /**
     * Adds middlewares BEFORE all the mountable objects and static-serve middlewares
     * @param {string} path
     * @param {TMiddleware<any>[]} middlewares
     */
    before(path: string, middlewares: TMiddleware<any>[]): HttpApplication;
    /**
     * Adds middlewares AFTER all the mountable objects and static-serve middlewares
     * @param {TMiddleware<any>[]} middlewares
     */
    after(middlewares: TMiddleware<any>[]): HttpApplication;
    /**
     * Adds middlewares AFTER all the mountable objects and static-serve middlewares
     * @param {string} path
     * @param {TMiddleware<any>[]} middlewares
     */
    after(path: string, middlewares: TMiddleware<any>[]): HttpApplication;
    /**
     * @param {string} path
     * @param {'before'|'after'} type
     * @param {TMiddleware<any>[]} middlewares
     * @protected
     */
    protected addMiddlewares(path: string, type: 'before' | 'after', middlewares: TMiddleware<any>[]): void;
    /**
     * @param {string} webRoot
     * @param {IStaticServe[]} serve
     */
    static(webRoot: string, serve: IStaticServe[]): HttpApplication;
    /**
     * @param {IInitializable[]} initializers
     * @returns {HttpApplication}
     */
    initializers(initializers: IInitializable[]): HttpApplication;
    /**
     * Launches the application
     * @returns {Promise<void>}
     */
    launch(): Promise<void>;
    protected initInitializables(): Promise<void>;
    protected initMiddlewares(type: 'before' | 'after'): Promise<void>;
    protected initStatic(): Promise<void>;
    protected initMountable(): Promise<void>;
}
export {};
