"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HttpApplication = void 0;
const path = require("path");
const express = require("express");
class HttpApplication {
    /**
     * @param {e.Application} express
     */
    constructor(express) {
        this.mountList = new Map();
        this.middlewaresBefore = new Map();
        this.middlewaresAfter = new Map();
        this.staticList = new Map();
        this.initializableList = new Set();
        this.expressApp = express;
    }
    /**
     * Mounts mountable objects to web-path
     * @param {string} path
     * @param {IMountable[]} mountable
     * First element - web-path to which mountable will be mount, second element - mountable object.
     * For example:
     *  ['/test', new TestMountable()]
     * @returns {HttpApplication}
     */
    mount(path, mountable) {
        if (!this.mountList.has(path))
            this.mountList.set(path, new Set());
        const set = this.mountList.get(path);
        if (set)
            mountable.forEach(m => set.add(m));
        return this;
    }
    before(pathOrMiddlewares, middlewares) {
        if (typeof pathOrMiddlewares !== 'string' && Array.isArray(pathOrMiddlewares))
            return this.before('*', pathOrMiddlewares);
        if (!middlewares)
            middlewares = [];
        this.addMiddlewares(pathOrMiddlewares, 'before', middlewares);
        return this;
    }
    after(pathOrMiddlewares, middlewares) {
        if (typeof pathOrMiddlewares !== 'string' && Array.isArray(pathOrMiddlewares))
            return this.after('*', pathOrMiddlewares);
        if (!middlewares)
            middlewares = [];
        this.addMiddlewares(pathOrMiddlewares, 'after', middlewares);
        return this;
    }
    /**
     * @param {string} path
     * @param {'before'|'after'} type
     * @param {TMiddleware<any>[]} middlewares
     * @protected
     */
    addMiddlewares(path, type, middlewares) {
        const mapName = type === 'before' ? 'middlewaresBefore' : 'middlewaresAfter';
        if (!this[mapName].has(path))
            this[mapName].set(path, new Set());
        const set = this[mapName].get(path);
        if (set)
            middlewares.forEach(m => set.add(m));
    }
    /**
     * @param {string} webRoot
     * @param {IStaticServe[]} serve
     */
    static(webRoot, serve) {
        if (!this.staticList.has(webRoot))
            this.staticList.set(webRoot, new Set());
        const set = this.staticList.get(webRoot);
        if (set)
            serve.forEach(fallback => set.add(fallback));
        return this;
    }
    /**
     * @param {IInitializable[]} initializers
     * @returns {HttpApplication}
     */
    initializers(initializers) {
        initializers.forEach(initializer => this.initializableList.add(initializer));
        return this;
    }
    /**
     * Launches the application
     * @returns {Promise<void>}
     */
    async launch() {
        await this.initInitializables();
        await this.initMiddlewares('before');
        await this.initStatic();
        await this.initMountable();
        await this.initMiddlewares('after');
    }
    //region Initialization of an application (after launch method is called)
    async initInitializables() {
        for (const initializable of this.initializableList.values()) {
            const promise = initializable.initialize();
            if (!!promise) {
                try {
                    await promise;
                }
                catch (e) {
                    continue;
                }
            }
        }
    }
    async initMiddlewares(type) {
        const mapName = type === 'before' ? 'middlewaresBefore' : 'middlewaresAfter';
        this[mapName].forEach((set, path) => {
            set.forEach(middleware => {
                const cArgs = [];
                if (path !== '*')
                    cArgs.push(path);
                cArgs.push(middleware);
                //@ts-ignore
                this.expressApp.use.apply(this.expressApp, cArgs);
            });
        });
    }
    async initStatic() {
        this.staticList.forEach((set, webRoot) => {
            set.forEach(serve => {
                //@ts-ignore
                this.expressApp.use(webRoot, express.static(serve[0], serve[1] || null));
            });
        });
    }
    async initMountable() {
        const initializeLater = [];
        this.mountList.forEach((set, webPath) => {
            set.forEach(mountable => {
                this.expressApp.use(path.posix.join(webPath, mountable[0]), mountable[1].router);
                initializeLater.push(mountable[1]);
            });
        });
        for (const mountable of initializeLater) {
            try {
                const promise = mountable.initialize();
                if (!!promise)
                    await promise;
            }
            catch (e) {
            }
        }
    }
}
exports.HttpApplication = HttpApplication;
