import {Application} from 'express';
import {IMountable} from './IMountable';
import {ServeStaticOptions} from 'serve-static';
import {IInitializable} from './IInitializable';
import * as path from 'path';
import express = require('express');
import {TMiddleware} from './TMiddleware';

interface IStaticServe {
    [0]: string;
    [1]?: ServeStaticOptions;
}

interface IMountPoint {
    /**
     * Mount path
     */
    [0]: string;
    /**
     * Mountable object
     */
    [1]: IMountable;
}

export class HttpApplication {

    protected readonly expressApp: Application;
    protected readonly mountList = new Map<string, Set<IMountPoint>>();
    protected readonly middlewaresBefore = new Map<string, Set<TMiddleware<any>>>();
    protected readonly middlewaresAfter = new Map<string, Set<TMiddleware<any>>>();
    protected readonly staticList = new Map<string, Set<IStaticServe>>();
    protected readonly initializableList = new Set<IInitializable>();

    /**
     * @param {e.Application} express
     */
    constructor(express: Application) {
        this.expressApp = express;
    }

    /**
     * Mounts mountable objects to web-path
     * @param {string} path
     * @param {IMountable[]} mountable
     * First element - web-path to which mountable will be mount, second element - mountable object.
     * For example:
     *  ['/test', new TestMountable()]
     * @returns {HttpApplication}
     */
    public mount(path: string, mountable: IMountPoint[]): HttpApplication {
        if (!this.mountList.has(path))
            this.mountList.set(path, new Set<IMountPoint>());

        const set = this.mountList.get(path);
        if(set)
            mountable.forEach(m => set.add(m));

        return this;
    }

    /**
     * Adds middlewares BEFORE all the mountable objects and static-serve middlewares
     * @param {TMiddleware<any>[]} middlewares
     */
    public before(middlewares: TMiddleware<any>[]): HttpApplication
    /**
     * Adds middlewares BEFORE all the mountable objects and static-serve middlewares
     * @param {string} path
     * @param {TMiddleware<any>[]} middlewares
     */
    public before(path: string, middlewares: TMiddleware<any>[]): HttpApplication
    public before(pathOrMiddlewares: string | TMiddleware<any>[], middlewares?: TMiddleware<any>[]): HttpApplication {
        if (typeof pathOrMiddlewares !== 'string' && Array.isArray(pathOrMiddlewares))
            return this.before('*', pathOrMiddlewares);

        if(!middlewares)
            middlewares = [];

        this.addMiddlewares(pathOrMiddlewares, 'before', middlewares);
        return this;
    }

    /**
     * Adds middlewares AFTER all the mountable objects and static-serve middlewares
     * @param {TMiddleware<any>[]} middlewares
     */
    public after(middlewares: TMiddleware<any>[]): HttpApplication
    /**
     * Adds middlewares AFTER all the mountable objects and static-serve middlewares
     * @param {string} path
     * @param {TMiddleware<any>[]} middlewares
     */
    public after(path: string, middlewares: TMiddleware<any>[]): HttpApplication
    public after(pathOrMiddlewares: string | TMiddleware<any>[], middlewares?: TMiddleware<any>[]): HttpApplication {
        if (typeof pathOrMiddlewares !== 'string' && Array.isArray(pathOrMiddlewares))
            return this.after('*', pathOrMiddlewares);

        if(!middlewares)
            middlewares = [];

        this.addMiddlewares(pathOrMiddlewares, 'after', middlewares);
        return this;
    }

    /**
     * @param {string} path
     * @param {'before'|'after'} type
     * @param {TMiddleware<any>[]} middlewares
     * @protected
     */
    protected addMiddlewares(path: string, type: 'before' | 'after', middlewares: TMiddleware<any>[]): void {
        const mapName = type === 'before' ? 'middlewaresBefore' : 'middlewaresAfter'
        if (!this[mapName].has(path))
            this[mapName].set(path, new Set<TMiddleware<any>>());

        const set = this[mapName].get(path);
        if(set)
            middlewares.forEach(m => set.add(m));
    }

    /**
     * @param {string} webRoot
     * @param {IStaticServe[]} serve
     */
    public static(webRoot: string, serve: IStaticServe[]): HttpApplication {
        if (!this.staticList.has(webRoot))
            this.staticList.set(webRoot, new Set<IStaticServe>());

        const set = this.staticList.get(webRoot);
        if(set)
            serve.forEach(fallback => set.add(fallback));
        return this;
    }

    /**
     * @param {IInitializable[]} initializers
     * @returns {HttpApplication}
     */
    public initializers(initializers: IInitializable[]): HttpApplication {
        initializers.forEach(initializer => this.initializableList.add(initializer));
        return this;
    }

    /**
     * Launches the application
     * @returns {Promise<void>}
     */
    public async launch(): Promise<void> {
        await this.initInitializables();
        await this.initMiddlewares('before');
        await this.initStatic();
        await this.initMountable();
        await this.initMiddlewares('after');
    }

    //region Initialization of an application (after launch method is called)

    protected async initInitializables(): Promise<void> {
        for (const initializable of this.initializableList.values()) {
            const promise = initializable.initialize();
            if (!!promise) {
                try {
                    await promise;
                } catch (e) {
                    continue;
                }
            }
        }
    }

    protected async initMiddlewares(type: 'before' | 'after'): Promise<void> {
        const mapName = type === 'before' ? 'middlewaresBefore' : 'middlewaresAfter';

        this[mapName].forEach((set, path) => {
            set.forEach(middleware => {
                const cArgs = [];
                if (path !== '*')
                    cArgs.push(path);

                cArgs.push(middleware);
                //@ts-ignore
                this.expressApp.use.apply(this.expressApp, cArgs);
            });
        });
    }

    protected async initStatic(): Promise<void> {
        this.staticList.forEach((set, webRoot) => {
            set.forEach(serve => {
                //@ts-ignore
                this.expressApp.use(webRoot, express.static(serve[0], serve[1] || null));
            });
        });
    }

    protected async initMountable(): Promise<void> {
        const initializeLater: IMountable[] = [];
        this.mountList.forEach((set, webPath) => {
            set.forEach(mountable => {
                this.expressApp.use(path.posix.join(webPath, mountable[0]), mountable[1].router);
                initializeLater.push(mountable[1]);
            });
        })

        for (const mountable of initializeLater) {
            try {
                const promise = mountable.initialize();
                if (!!promise)
                    await promise;
            } catch (e) {
            }
        }
    }
    //endregion

}
